##############################################
package aims2egroups;
##############################################
#
# The E-Groups Class.  
#
# Author: Dan Dengate <dan.dengate@cern.ch>
#
##############################################

use strict;
use LWP; # Debug?

my $logger;

1;

##############################################
sub new
##############################################
{
   my($class, $dbh, $user, %conf) = @_;
   my($self) = bless{}, $class;
   $self->initialize($dbh, $user, %conf);
   return $self;
}

##############################################
sub initialize
##############################################
{
   my($self, $dbh, $user, %conf) = @_;
   
   $self->{_DB} = $dbh;
   $self->{_USER} = $user;
   %{$self->{_CONF}} = %conf;

   $logger = aims2log->new($dbh, $user);   
   
   my $win_ep 		= $conf{'WIN_EP'};
   unless($win_ep){
      die "Error: Windows End Point is undefined in configuration.\n";
   }
   my $win_realm 	= $conf{'WIN_REALM'};
   unless($win_realm){
      die "Error: Windows Realm is undefined in configuration.\n";
   }
   my $win_user 	= $conf{'WIN_USER'};
   unless($win_user){
      die "Error: Windows user undefined in configuration.\n";
   }
   my $win_pass 	= $conf{'WIN_PASS'};
   unless($win_pass){
      die "Error: Windows password undefined in configuration.\n";
   }
 
   unless($win_ep && $win_realm && $win_user && $win_pass){
      die "Error: Failed to load configuration for e-groups.\n";
   } 

   # Establish a Secure connection to egroups.
   my $ua = new LWP::UserAgent;
      $ua->agent("AgentName/0.1 " . $ua->agent);
      $ua->credentials($win_ep, $win_realm, $win_user, $win_pass);
   unless($ua){
      $logger->("e-groups error: Failed to establish connection.");
      die "Error: Cannot establish connection to e-groups.\n";
   }
   $self->{_EGROUPS_UA} = $ua;
}

##############################################
sub member_of_group
##############################################
{
   my($self, @groups) = @_;
   
   unless(@groups){
     die "Error: Member of Group failed. No groups provided.\n";
   }
   
   my %conf = %{$self->{_CONF}};
   my $user = $self->{_USER};
   my $ua = $self->{_EGROUPS_UA};
   my $is_member_of_group;
   
   foreach my $group (@groups) {
      chomp($group);
      my $req = HTTP::Request->new(POST=>$conf{'EGROUP_URL'});
      $req->content_type('application/x-www-form-urlencoded');
      $req->content("ListName=$group");
      my $res = $ua->request($req);
      unless($res){
	 $logger->log("Error: Request to egroups failed.");
         die "Error: Request to e-groups failed.\n";
      }
      if($res->is_success) {
         my $re = qr{<string>(.+?)</string>}s;
         $res->content =~ $re;
         if($res->content =~ /$user/) {
            $is_member_of_group = "Success";
            last;
         }    
      } else {
         die "Error: Failed to get results from e-groups.\n";
      }
      next;
   }
   return $is_member_of_group;
}

##############################################
sub translate_email_to_login
##############################################
{
   # Translate "daniel.robert.dengate@cern.ch" to NICE login "dengate"
   my($self, @emails) = @_;
   unless(@emails){ 
      die "Error: Login translation failed. No input provided.\n";
   }
   my %conf = %{$self->{_CONF}};
   my $ua = $self->{_EGROUPS_UA};
   my $is_member_of_group;
   my @users; 
   my $win_url = $conf{'GETLOGIN_URL'};
   unless($win_url){
      die "Error: Cannot get logins. Error with connection string.\n";
   }
   for my $email (@emails){
      my $req = HTTP::Request->new(POST=>$win_url);      
      $req->content_type('application/x-www-form-urlencoded');
      $req->content("username=$email");
      my $res = $ua->request($req);
      unless($res){
         die "Error: Failed request to Windows Authentication Services.\n";
      }
      if($res->is_success) {
         my $re = qr{<login>(.+?)</login>}s;
         $res->content =~ $re;
         my $login = $1;
         push(@users, $login)
      } else {
         die "Error: No success getting NICE login.\n";
      } 
   }   
   @users;
}

##############################################
sub translate_owner_to_egroup
##############################################
{
   my($self, @owners) = @_;
   unless(@owners){
      die "Error: Cannot translate owner to group. Owners Null.\n";
   }

   # FixMe: These should really be in the database somewhere.
   my %owner_hack = %aims2conf::owner_hack;
   my @groups;
   for my $owner (@owners){
      while( my($email, $egroups) = each(%owner_hack) ){
         if(lc($owner) eq lc($email)){
            push(@groups, split(",", $egroups));
         }
      }
   }   
   return @groups;
}
